/*
 * ----------------------------------------------------------------------------
 *
 * zzzforward.c -- listen on a socket and forward incomming connections
 * Copyright (C) 2003 Fabian "zzznowman" Pietsch
 * Version 0.1.2 (2003-12-27)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * ----------------------------------------------------------------------------
 *
 */

#include <errno.h>	/* errno */
#include <err.h>	/* warn() */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>	/* read(), write(), sleep() */
#include <fcntl.h>	/* fcntl() */

#include <netdb.h>	/* gethostbyname() */
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>	/* inet_aton() */


/* retries on send queue full */
#define TRIES_MAX 10	/* maximum number of retries when send queue is full */
#define TRIES_DELAY 50	/* delay between retries (ms)                        */

/* structure holding connection informations (linked list) */
struct zf_conn {
	int id;				/* unique connection id                   */
	int s_client, s_server;		/* socket file descriptors                */
	struct sockaddr_in sa_client;	/* "client" which connected to us         */
	struct sockaddr_in sa_server;	/* "server" to which we connected in turn */
	struct zf_conn *next;		/* pointer to next element in linked list */
};

/* function prototypes */
int zf_accept(int listens, struct zf_conn **connendp);
int zf_connect(int *s, struct sockaddr_in *sa);
int zf_disconnect(struct zf_conn **conn);
int zf_listen(int *s, int port);
int zf_transmit(struct zf_conn *conn);

#if 0
char *desthost = "irc.ham.de.euirc.net";
char *destport = "6667";
#else
char *desthost = "localhost";
char *destport = "6667";
#endif
/* port zzzforward will be listening on */
int listenport = 7000;

/* active connections list */
struct zf_conn *connections;


/* accept incoming connections */
int zf_accept(int listens, struct zf_conn **connendp) {
	static int nextid;
	int s;
	struct sockaddr_in sa;
	size_t sa_len = sizeof(sa);
	struct zf_conn *conn;

	/* try accept()ing a connection */
	if ((s = accept(listens, &sa, &sa_len)) == -1) {
		/* no incoming connections on non-blocking I/O? */
		if (errno == EAGAIN)
			/* success (but no new connection) */
			return(0);

		/* a real error occured */
		perror("error: accept()");
		return(-1);
	}
	if (sa_len != sizeof(sa)) {
		fprintf(stderr, "error: accept() returned unknown client data structure of size %d\n", sa_len);
		close(s);
		return(-1);
	}

	/* set new connection non-blocking */
	if (fcntl(s, F_SETFL, fcntl(s, F_GETFL) | O_NONBLOCK)) {
		perror("error: fcntl()");
		close(s);
		return(-1);
	}

	/* allocate new connection information data structure */
	if ((conn = malloc(sizeof(struct zf_conn))) == NULL) {
		perror("error: malloc()");
		close(s);
		return(-1);
	}

	/* fill in connection information */
	conn->id = nextid++;
	conn->s_client = s;
	conn->sa_client = sa;
	conn->next = NULL;

	/* connect to real server */
	if (zf_connect(&conn->s_server, &conn->sa_server)) {
		fprintf(stderr, "error: connecting to real server failed -- dropping connection %d\n", conn->id);
		close(s);
		free(conn);
	}

	/* append connection information to list of active connections */
	*connendp = conn;

	/* log what happened */
	printf("new connection %d (%s:%d", conn->id,
	  inet_ntoa(conn->sa_client.sin_addr),
	  ntohs(conn->sa_client.sin_port));
	printf(" -> %s:%d)\n",
	  inet_ntoa(conn->sa_server.sin_addr),
	  ntohs(conn->sa_server.sin_port));

	/* success */
	return(0);
}

/* connect to desthost:destport */
int zf_connect(int *s, struct sockaddr_in *sa) {
	struct hostent *he;

	/* lookup destination host */
	if ((he = gethostbyname(desthost)) == NULL) {
		fprintf(stderr, "error: gethostbyname() failed (%s)\n",
		  (h_errno == HOST_NOT_FOUND) ? "host not found"
		  : "unknown error");
		return(-1);
	}

	/* create a socket */
	if ((*s = socket(PF_INET, SOCK_STREAM, 0)) == -1) {
		perror("error: socket()");
		return(-1);
	}

	/* prepare sa */
	sa->sin_family = AF_INET;
	sa->sin_port = htons(atol(destport));
	sa->sin_addr = *((struct in_addr *)he->h_addr_list[0]);

	/* connect the socket */
	if (connect(*s, sa, sizeof(*sa))) {
		perror("error: connect()");
		close(*s);
		return(-1);
	}

	/* set s nonblocking */
	if (fcntl(*s, F_SETFL, fcntl(*s, F_GETFL) | O_NONBLOCK)) {
		perror("error: fcntl()");
		close(*s);
		return(-1);
	}

	/* success */
	return(0);
}

/* disconnect an active connection */
int zf_disconnect(struct zf_conn **conn) {
	printf("[%d] disconnecting\n", (*conn)->id);

	/* close connection to real server */
	if (close((*conn)->s_server))
		warn("error: [%d] close() on server socket", (*conn)->id);
		/* return(-1); */

	/* close connection to client */
	if (close((*conn)->s_client))
		warn("error: [%d] close() on client socket", (*conn)->id);

	/* free active connection data structure */
	free(*conn);

	/* finalize linked list */
	*conn = NULL;

	/* success */
	return(0);
}

/* perform steps necessary for accepting incoming connections */
int zf_listen(int *s, int port) {
	struct sockaddr_in sa;

	/* create a socket */
	if ((*s = socket(PF_INET, SOCK_STREAM, 0)) == -1) {
		perror("error: socket()");
		return(-1);
	}

	/* set s nonblocking
	 * this will not be inherited by accept()ed sockets,
	 * but accept() won't block
	 */
	if (fcntl(*s, F_SETFL, fcntl(*s, F_GETFL) | O_NONBLOCK)) {
		perror("error: fcntl()");
		close(*s);
		return(-1);
	}

	/* prepare sa */
	sa.sin_family = AF_INET;
	sa.sin_port = htons(port);
	sa.sin_addr.s_addr = INADDR_ANY;

	/* bind on specified port */
	if (bind(*s, &sa, sizeof(sa))) {
		perror("error: bind()");
		close(*s);
		return(-1);
	}

	/* finally perform listen() call on the socket */
	if (listen(*s, 3)) {
		perror("error: listen()");
		close(*s);
		return(-1);
	}

	/* success */
	return(0);
}

/* perform one transfer cycle between two file descriptors */
int zf_transmit(struct zf_conn *conn) {
	char buffer[65535], *bufp;
	int incount, outcount, tries;
	int  *fdinp  = &conn->s_server,
	     *fdoutp = &conn->s_client;

__PASSDATA__:
	switch (incount = read(*fdinp, buffer, sizeof(buffer))) {
	case -1: /* error */
		/* no data available on non-blocking IO? */
		if (errno != EAGAIN)
			/* other error => fail */
			return(-1);

		/* skip this transfer direction (lack of data) */
		break;
	case  0: /* end of file */
		/* fail */
		return(-1);
	default: /* received some data */
		/* try sending data until everything is sent */
		bufp = buffer; tries = 0;
		while ((outcount = write(*fdoutp, bufp, incount)) < incount) {
			switch (outcount) {
			case -1: /* error */
				/* no buffer space available on non-blocking IO? */
				if (errno != EAGAIN)
					/* other error => fail */
					return(-1);

				/* TRIES_MAX exceeded? */
				if (++tries > TRIES_MAX) {
					/* timeout */
					fprintf(stderr, "error: [%d] %s timeout\n", conn->id,
					  (*fdoutp == conn->s_server) ? "server" : "client");
					return(-1);
				}

				/* delay for TRIES_DELAY milliseconds,
				 * then try again until some "real" error occurs
				 */
				usleep(TRIES_DELAY);
				continue;
			case  0: /* "nothing was written" => ? */
			default: /* sent some data */
				bufp += outcount;
				incount -= outcount;
			}
		}
	}

	/* first run? */
	if (*fdinp == conn->s_server) {
		/* start over again with source and destination exchanged */
		fdinp  = &conn->s_client;
		fdoutp = &conn->s_server;
		goto __PASSDATA__;
	}

	/* success */
	return(0);
}

int main(int argc, char **argv) {
	int listens;	/* listening socket */
	struct zf_conn *conn, **connendp;

	/* listen on listenport */
	if (zf_listen(&listens, listenport))
		exit(EXIT_FAILURE);

	/* enter main loop */
	while (1) {
		/* walk through all active connections */
		conn = *(connendp = &connections);
		while (conn) {
			/* perform a single transfer cycle */
			if (zf_transmit(conn)) {
				/* error => close connection */
				fprintf(stderr, "[%d] transmission error -- dropping connection\n", conn->id);
				zf_disconnect(connendp);
			}

			/* step ahead in linked list */
			conn = *(connendp = &conn->next);
		}

		/* accept new connections */
		if (zf_accept(listens, connendp)) {
			fprintf(stderr, "error: can't accept() -- exiting\n");
			exit(EXIT_FAILURE);
		}

#if 0
		/* connect to destination host */
		if (zf_connect(&s, desthost, destport))
			exit(EXIT_FAILURE);
#endif

		/* don't consume ~100% CPU time => sleep 10ms */
		usleep(10);
	}

	/* success (well, currently won't ever be reached.. *hide*) */
	exit(EXIT_SUCCESS);
}
